/* $Header: /usr/home/yav/fkiss/RCS/dir.c,v 1.13 1998/02/03 00:36:13 yav Exp $
 * Simple directory access routine
 * written by yav (UHD98984@biglobe.ne.jp)
 *
 * #define USE_BSD_DIRLIB 1
 *  use opendir, readdir and closedir library routines.
 * #define USE_BSD_DIRLIB 0
 *  use popen to execute "ls -af".
 *
 * #define DIR_STANDALONE_TEST
 * for self checking
 *
 */

char id_dir[] = "$Id: dir.c,v 1.13 1998/02/03 00:36:13 yav Exp $";

#include <stdio.h>

#ifdef DIR_STANDALONE_TEST
# include "config.h"
#else
# include <X11/Xlib.h>
# include "config.h"
# include "headers.h"
# include "fkiss.h"
# define PUBLIC_DIR_C
# include "extern.h"
#endif

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif


#ifdef DIR_STANDALONE_TEST
#define ks_malloc malloc
#define ks_realloc realloc
#define ks_strdup strdup
extern char *malloc();
extern char *realloc();
extern void free();
extern char *strdup();
extern char *strncpy();
#endif

#ifndef MAXPATH
# ifdef MAXPATHLEN
#  define MAXPATH MAXPATHLEN
# else
#  define MAXPATH 1024
# endif
#endif

#ifndef USE_BSD_DIRLIB
#warning USE_BSD_DIRLIB not defined!
#endif

/* dir_ls - get filename list in directory
 * return filename pointer list (NULL : directory open error)
 */
char **dir_ls(path, func)
     char *path;		/* directory */
     int (*func)();		/* filename matching function (NULL:all) */
{
  int r;
  int i;
  char **list;
#if USE_BSD_DIRLIB
  DIR *dp;
  struct dirent *p;
#else
  FILE *fp;
#endif
  char buf[MAXPATH];
  
  r = 0;
  list = NULL;
#if USE_BSD_DIRLIB
  if ((dp = opendir(path)) != NULL) {
    list = (char **)ks_malloc(sizeof(char *));
    while ((p = readdir(dp)) != NULL) {
      i = NAMLEN(p);
      strncpy(buf, p->d_name, i);
      buf[i] = '\0';
      if (func == NULL || (*func)(buf)) {
	*(list+r++) = ks_strdup(buf);
	list = (char **)ks_realloc(list, sizeof(char *) * (r+1));
      }
    }
    *(list+r) = NULL;
    closedir(dp);
  }
#else /* USE_BSD_DIRLIB */
  sprintf(buf, "ls -af %s", path);
  if ((fp = popen(buf, "r")) != NULL) {
    list = (char **)ks_malloc(sizeof(char *));
    while (fgets(buf, sizeof(buf), fp) != NULL) {
      i = strlen(buf) - 1;
      if (i >= 0 && buf[i] == '\n')
	buf[i] = '\0';		/* cut '\n' */
      if (func == NULL || (*func)(buf)) {
	*(list+r++) = ks_strdup(buf);
	list = (char **)ks_realloc(list, sizeof(char *) * (r+1));
      }
    }
    *(list+r) = NULL;
    pclose(fp);
  }
#endif /* USE_BSD_DIRLIB */
  return list;
}

/* free all allocated filename list */
void dir_free(list)
     char **list;
{
  char **p;
  
  if (list != NULL) {
    for (p = list; *p != NULL; p++)
      free(*p);
    free(list);
  }    
}

#ifdef DIR_STANDALONE_TEST

int cmpfunc(char *name)
{
  return (*name == 'f');
}

void main(argc, argv)
     int argc;
     char **argv;
{
  int i;
  char **buf;
  
  if (argc < 2)
    exit(1);
#if USE_BSD_DIRLIB
  printf("* dir lib *\n");
#else
  printf("* dir ls  *\n");
#endif
  buf = dir_ls(*++argv, cmpfunc);
  if (buf == NULL) {
    fprintf(stderr, "``%s'' read error!\n", *argv);
    exit(1);
  }
  for (i = 0; *(buf+i) != NULL; i++) {
    printf("[%s]\n", *(buf+i));
  }
  dir_free(buf);
  printf("Total %d files\n", i);
  exit(0);
}

#endif /* DIR_STANDALONE_TEST */

/* End of file */
